from __future__ import annotations

import asyncio
import json
import logging
import os
import sys
from asyncio import Queue
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import click
import click_logging
import tabulate
import tqdm
from pystac import Asset, Item, ItemCollection

from . import Config, ErrorStrategy, _functions
from .client import Clients
from .config import DEFAULT_S3_MAX_ATTEMPTS, DEFAULT_S3_RETRY_MODE
from .errors import DownloadError
from .messages import (
    ErrorAssetDownload,
    FinishAssetDownload,
    OpenUrl,
    SkipAssetDownload,
    StartAssetDownload,
    WriteChunk,
)
from .types import MessageQueue

logger = logging.getLogger(__name__)
click_logging.basic_config(logger)

# Needed until we drop Python 3.8
if TYPE_CHECKING:
    Tqdm = tqdm.tqdm[Any]
else:
    Tqdm = tqdm.tqdm


@click.group()
@click_logging.simple_verbosity_option(logger)  # type: ignore
def cli() -> None:
    """Work with STAC assets.

    See each subcommand's help text for more information:

        $ stac-asset download --help
        $ stac-asset info --help
    """


@cli.command()
@click.argument("href", required=False)
@click.argument("directory", required=False)
@click.option(
    "-a",
    "--alternate-assets",
    help="Alternate asset hrefs to prefer, if available",
    multiple=True,
)
@click.option("-i", "--include", help="Asset keys to include", multiple=True)
@click.option(
    "-x",
    "--exclude",
    help="Asset keys to exclude (can't be used with include)",
    multiple=True,
)
@click.option("-f", "--file-name", help="The output file name")
@click.option(
    "-q",
    "--quiet",
    help="Do not print anything to standard output.",
    default=False,
    is_flag=True,
    show_default=True,
)
@click.option(
    "--s3-requester-pays",
    help="If downloading via the s3 client, enable requester pays",
    default=False,
    is_flag=True,
    show_default=True,
)
@click.option(
    "--s3-retry-mode",
    help="If downloading via the s3 client, the retry mode (standard, legacy, and "
    "adaptive)",
    default=DEFAULT_S3_RETRY_MODE,
)
@click.option(
    "--s3-max-attempts",
    help="If downloading via the s3 client, the max number of retries",
    default=DEFAULT_S3_MAX_ATTEMPTS,
)
@click.option(
    "-k",
    "--keep",
    help=(
        "If warning on error, keep assets that couldn't be downloaded with their "
        "original hrefs. If false, delete those assets from the item."
    ),
    default=False,
    is_flag=True,
    show_default=True,
)
@click.option(
    "--fail-fast",
    help="Fail immediately on download error, instead of waiting until all are "
    "complete.",
    default=False,
    is_flag=True,
    show_default=True,
)
@click.option(
    "--overwrite",
    help="Overwrite existing files if they exist on the filesystem",
    default=False,
    is_flag=True,
    show_default=True,
)
# TODO add option to disable content type checking
def download(
    href: Optional[str],
    directory: Optional[str],
    alternate_assets: List[str],
    include: List[str],
    exclude: List[str],
    file_name: Optional[str],
    quiet: bool,
    s3_requester_pays: bool,
    s3_retry_mode: str,
    s3_max_attempts: int,
    keep: bool,
    fail_fast: bool,
    overwrite: bool,
) -> None:
    """Download STAC assets from an item or item collection.

    If href is not provided, or is ``-``, the item or item collection is parsed
    as JSON from standard input. If the directory is not provided, the current
    working directory is used.

    These three examples are equivalent, and download the assets to the current
    working directory:

        $ stac-asset download item.json

        $ stac-asset download item.json .

        $ cat item.json | stac-asset download

    To only include certain asset keys:

        $ stac-asset download -i asset-key-to-include item.json
    """
    asyncio.run(
        download_async(
            href,
            directory,
            alternate_assets,
            include,
            exclude,
            file_name,
            quiet,
            s3_requester_pays,
            s3_retry_mode,
            s3_max_attempts,
            keep=keep,
            fail_fast=fail_fast,
            overwrite=overwrite,
        )
    )


async def download_async(
    href: Optional[str],
    directory: Optional[str],
    alternate_assets: List[str],
    include: List[str],
    exclude: List[str],
    file_name: Optional[str],
    quiet: bool,
    s3_requester_pays: bool,
    s3_retry_mode: str,
    s3_max_attempts: int,
    keep: bool,
    fail_fast: bool,
    overwrite: bool,
) -> None:
    config = Config(
        alternate_assets=alternate_assets,
        include=include,
        exclude=exclude,
        s3_requester_pays=s3_requester_pays,
        s3_retry_mode=s3_retry_mode,
        s3_max_attempts=s3_max_attempts,
        error_strategy=ErrorStrategy.KEEP if keep else ErrorStrategy.DELETE,
        warn=not fail_fast,
        fail_fast=fail_fast,
        overwrite=overwrite,
    )

    input_dict = await read_as_dict(href, config)
    if directory is None:
        directory_str = os.getcwd()
    else:
        directory_str = str(directory)
    if quiet:
        messages = None
    else:
        messages = Queue()

    type_ = input_dict.get("type")
    if type_ is None:
        if not quiet:
            print("ERROR: missing 'type' field on input dictionary", file=sys.stderr)
        sys.exit(1)
    elif type_ == "Feature":
        item = Item.from_dict(input_dict)
        if href and href != "-":
            item.set_self_href(href)
            item.make_asset_hrefs_absolute()

        async def download() -> Union[Item, ItemCollection]:
            return await _functions.download_item(
                item,
                directory_str,
                file_name=file_name,
                infer_file_name=False,
                config=config,
                messages=messages,
            )

    elif type_ == "FeatureCollection":
        item_collection = ItemCollection.from_dict(input_dict)

        async def download() -> Union[Item, ItemCollection]:
            return await _functions.download_item_collection(
                item_collection,
                directory_str,
                file_name=file_name,
                config=config,
                messages=messages,
            )

    else:
        if not quiet:
            print(f"ERROR: unsupported 'type' field: {type_}", file=sys.stderr)
        sys.exit(2)

    task = asyncio.create_task(report_progress(messages))
    try:
        output = await download()
    except DownloadError:
        sys.exit(1)

    if messages:
        await messages.put(None)
    await task

    if not quiet:
        if file_name:
            print(f"Output STAC JSON written to {Path(directory_str) / file_name}")
        else:
            json.dump(output.to_dict(transform_hrefs=False), sys.stdout)


async def read_as_dict(href: Optional[str], config: Config) -> Dict[str, Any]:
    if href is None or href == "-":
        data = json.load(sys.stdin)
    else:
        data = json.loads(await read(href, config))
    if not isinstance(data, dict):
        raise ValueError(f"input is not a dictionary: {data.__type__}")
    else:
        return data


async def read(href: str, config: Config) -> bytes:
    clients = Clients(config)
    async with await clients.get_client(href) as client:
        data = b""
        async for chunk in client.open_href(href):
            data += chunk
        return data


async def report_progress(messages: Optional[MessageQueue]) -> None:
    if messages is None:
        return
    progress_bar = tqdm.tqdm(
        unit="B",
        unit_scale=True,
        unit_divisor=1024,
    )
    sizes = dict()
    owners = dict()
    assets = 0
    done = 0
    errors = 0
    skips = 0
    total = 0
    n = 0
    progress_bar.set_postfix_str(f"{errors} errors")
    while True:
        message = await messages.get()
        if isinstance(message, StartAssetDownload):
            assets += 1
            if message.owner_id:
                owners[message.key] = message.owner_id
            progress_bar.set_description(f"{done}/{assets}")
        elif isinstance(message, OpenUrl):
            if message.size:
                total += message.size
                sizes[str(message.url)] = message.size
                progress_bar.reset(total=total)
                progress_bar.update(n)
        elif isinstance(message, SkipAssetDownload):
            skips += 1
            progress_bar.set_description_str(f"{done}/{assets}")
            progress_bar.set_postfix_str(f"{errors} errors, {skips} skips")
        elif isinstance(message, FinishAssetDownload):
            done += 1
            progress_bar.set_description_str(f"{done}/{assets}")
        elif isinstance(message, ErrorAssetDownload):
            done += 1
            errors += 1
            if message.href in sizes:
                total -= sizes[message.href]
                progress_bar.reset(total=total)
                progress_bar.update(n)
            progress_bar.set_postfix_str(f"{errors} errors, {skips} skips")
            progress_bar.set_description_str(f"{done}/{assets}")
            if message.key in owners:
                name = f"{owners[message.key]}[{message.key}]"
            else:
                name = f"[{message.key}]"
            progress_bar.write(f"ERROR: {name} - {message.error}", file=sys.stderr)
        elif isinstance(message, WriteChunk):
            n += message.size
            progress_bar.update(message.size)
        elif message is None:
            progress_bar.write("\n", file=sys.stderr)
            progress_bar.close()
            return


@dataclass
class Download:
    key: str
    item_id: Optional[str]
    href: str
    path: str
    progress_bar: Tqdm


@cli.command()
@click.argument("HREF", required=False)
@click.option(
    "-a",
    "--alternate-assets",
    help="Alternate asset hrefs to prefer, if available",
    multiple=True,
)
@click.option(
    "--s3-requester-pays",
    help="If checking via the s3 client, enable requester pays",
    default=False,
    is_flag=True,
    show_default=True,
)
@click.option(
    "--s3-retry-mode",
    help="If checking via the s3 client, the retry mode (standard, legacy, and "
    "adaptive)",
    default=DEFAULT_S3_RETRY_MODE,
)
@click.option(
    "--s3-max-attempts",
    help="If checking via the s3 client, the max number of retries",
    default=DEFAULT_S3_MAX_ATTEMPTS,
)
def info(
    href: Optional[str],
    alternate_assets: List[str],
    s3_requester_pays: bool,
    s3_retry_mode: str,
    s3_max_attempts: int,
) -> None:
    asyncio.run(
        info_async(
            href=href,
            alternate_assets=alternate_assets,
            s3_requester_pays=s3_requester_pays,
            s3_max_attempts=s3_max_attempts,
            s3_retry_mode=s3_retry_mode,
        )
    )


async def info_async(
    href: Optional[str],
    alternate_assets: List[str],
    s3_requester_pays: bool,
    s3_retry_mode: str,
    s3_max_attempts: int,
) -> None:
    """Prints information about an item or item collection.

    $ stac-asset info item.json
    """
    config = Config(
        alternate_assets=alternate_assets,
        s3_requester_pays=s3_requester_pays,
        s3_retry_mode=s3_retry_mode,
        s3_max_attempts=s3_max_attempts,
    )
    input_dict = await read_as_dict(href, config)
    type_ = input_dict.get("type")
    tasks = set()
    clients = Clients(config)
    if type_ is None:
        print("ERROR: missing 'type' field on input dictionary", file=sys.stderr)
        sys.exit(1)
    elif type_ == "Feature":
        item = Item.from_dict(input_dict)
        if href and href != "-":
            item.set_self_href(href)
            item.make_asset_hrefs_absolute()
        for key, asset in item.assets.items():
            tasks.add(asyncio.create_task(get_asset_info(key, asset, config, clients)))
    elif type_ == "FeatureCollection":
        item_collection = ItemCollection.from_dict(input_dict)
        for item in item_collection.items:
            for key, asset in item.assets.items():
                tasks.add(
                    asyncio.create_task(get_asset_info(key, asset, config, clients))
                )
    else:
        print(f"ERROR: invalid 'type' field: {type_}", file=sys.stderr)
        sys.exit(1)

    asset_infos = await asyncio.gather(*tasks)
    await clients.close_all()

    table_dict = dict()
    for asset_info in asset_infos:
        assert isinstance(asset_info, AssetInfo)
        if asset_info.key not in table_dict:
            table_dict[asset_info.key] = {
                "Key": asset_info.key,
                "Client": asset_info.client,
                "Media type": asset_info.media_type,
                "Exists": asset_info.exists,
                "Note": asset_info.note,
            }
        else:
            if asset_info.exists != table_dict[asset_info.key]["Exists"]:
                table_dict[asset_info.key]["Exists"] = "Sometimes"
            if asset_info.client != table_dict[asset_info.key]["Client"]:
                table_dict[asset_info.key]["Client"] = "Various"
            if asset_info.media_type != table_dict[asset_info.key]["Media type"]:
                table_dict[asset_info.key]["Media type"] = "Various"
            if asset_info.note and table_dict[asset_info.key]["Note"]:
                table_dict[asset_info.key]["Note"] = (
                    str(table_dict[asset_info.key]["Note"]) + "\n" + asset_info.note
                )

    keys = sorted(table_dict.keys())
    headers = ["Asset", "Client", "Exists", "Media type", "Note"]
    table_data = list()
    for key in keys:
        value = table_dict[key]
        table_data.append(
            [
                value["Key"],
                value["Client"],
                value["Exists"],
                value["Media type"],
                value["Note"],
            ]
        )
    print(tabulate.tabulate(table_data, headers=headers))


async def get_asset_info(
    key: str, asset: Asset, config: Config, clients: Clients
) -> AssetInfo:
    # TODO refactor with asset_exists
    href = _functions.get_absolute_asset_href(asset, config.alternate_assets)
    if href:
        client = await clients.get_client(href)
        try:
            await client.assert_href_exists(href)
        except Exception as error:
            note = str(error)
            exists = False
        else:
            note = ""
            exists = True
        # TODO clients should probably specify this explicitly
        name = type(client).__name__.lower()
        # We can't use `removesuffix` because it was added in Python 3.9
        if name.endswith("client"):
            name = name[: -len("client")]
        return AssetInfo(
            key=key, client=name, media_type=asset.media_type, exists=exists, note=note
        )
    else:
        return AssetInfo(
            key=key,
            client="n/a",
            exists=False,
            note="Could not make absolute href",
        )


@dataclass
class AssetInfo:
    key: str
    client: str
    exists: bool
    media_type: Optional[str] = None
    note: str = ""


if __name__ == "__main__":
    cli()



import pytest
from stac_asset import Config, HttpClient

pytestmark = [
    pytest.mark.network_access,
    pytest.mark.asyncio,
]


async def test_href_exists() -> None:
    async with await HttpClient.from_config(Config()) as client:
        assert await client.href_exists("https://github.com/stac-utils/stac-asset")



import json
import os.path
from asyncio import Queue
from pathlib import Path

import pytest
import stac_asset
from pystac import Asset, Collection, Item, ItemCollection
from pytest import MonkeyPatch
from stac_asset import (
    AssetOverwriteError,
    Config,
    ConfigError,
    DownloadError,
    DownloadWarning,
    ErrorStrategy,
    FileNameStrategy,
    S3Client,
)
from stac_asset.types import MessageQueue

pytestmark = [
    pytest.mark.asyncio,
]


async def test_download_item(tmp_path: Path, item: Item) -> None:
    item = await stac_asset.download_item(item, tmp_path, infer_file_name=False)
    assert os.path.exists(tmp_path / "20201211_223832_CS2.jpg")
    asset = item.assets["data"]
    assert asset.href == str(tmp_path / "20201211_223832_CS2.jpg")


async def test_download_item_with_file_name(tmp_path: Path, item: Item) -> None:
    await stac_asset.download_item(item, tmp_path, file_name="item.json")
    item = Item.from_file(str(tmp_path / "item.json"))
    assert item.assets["data"].href == "./20201211_223832_CS2.jpg"


async def test_download_missing_asset_error(tmp_path: Path, item: Item) -> None:
    item.assets["does-not-exist"] = Asset("not-a-file.md5")
    with pytest.raises(DownloadError):
        await stac_asset.download_item(item, tmp_path, config=Config(warn=False))


async def test_download_missing_asset_warn(tmp_path: Path, item: Item) -> None:
    item.assets["does-not-exist"] = Asset("not-a-file.md5")
    with pytest.warns(DownloadWarning):
        await stac_asset.download_item(item, tmp_path, config=Config(warn=True))


async def test_download_missing_asset_keep(
    tmp_path: Path, item: Item, data_path: Path
) -> None:
    item.assets["does-not-exist"] = Asset("not-a-file.md5")
    with pytest.warns(DownloadWarning):
        item = await stac_asset.download_item(
            item,
            tmp_path,
            config=Config(error_strategy=ErrorStrategy.KEEP, warn=True),
            infer_file_name=False,
        )
    assert item.assets["does-not-exist"].href == str(data_path / "not-a-file.md5")


async def test_download_missing_asset_delete(tmp_path: Path, item: Item) -> None:
    item.assets["does-not-exist"] = Asset("not-a-file.md5")
    with pytest.warns(DownloadWarning):
        item = await stac_asset.download_item(
            item,
            tmp_path,
            config=Config(error_strategy=ErrorStrategy.DELETE, warn=True),
        )
    assert "does-not-exist" not in item.assets


async def test_download_nonexistent_asset(tmp_path: Path, item: Item) -> None:
    # this previously had a bug where the code assumed the directory had
    # been created by downloading the assets when trying to write the item json,
    # but it hadn't, so a failure occurred
    await stac_asset.download_item(
        item,
        tmp_path / "dir-that-doesnt-exist",
        file_name="item.json",
        config=Config(include=["non-existent-asset"]),
    )


async def test_download_missing_asset_fail_fast(tmp_path: Path, item: Item) -> None:
    item.assets["does-not-exist"] = Asset("not-a-file.md5")
    with pytest.raises(FileNotFoundError):
        await stac_asset.download_item(
            item,
            tmp_path,
            config=Config(fail_fast=True),
        )


async def test_download_item_collection(
    tmp_path: Path, item_collection: ItemCollection
) -> None:
    item_collection = await stac_asset.download_item_collection(
        item_collection,
        tmp_path,
        file_name=None,
    )
    assert os.path.exists(tmp_path / "test-item" / "20201211_223832_CS2.jpg")
    asset = item_collection.items[0].assets["data"]
    assert asset.href == str(tmp_path / "test-item/20201211_223832_CS2.jpg")


async def test_download_item_collection_with_file_name(
    tmp_path: Path, item_collection: ItemCollection
) -> None:
    await stac_asset.download_item_collection(
        item_collection, tmp_path, file_name="item-collection.json"
    )
    item_collection = ItemCollection.from_file(str(tmp_path / "item-collection.json"))
    assert item_collection.items[0].assets["data"].href == str(
        tmp_path / "test-item/20201211_223832_CS2.jpg"
    )


async def test_download_item_collection_with_file_name_and_cd(
    tmp_path: Path, item_collection: ItemCollection, monkeypatch: MonkeyPatch
) -> None:
    monkeypatch.chdir(tmp_path)
    await stac_asset.download_item_collection(
        item_collection, ".", file_name="item-collection.json"
    )
    item_collection = ItemCollection.from_file(str(tmp_path / "item-collection.json"))
    assert item_collection.items[0].assets["data"].href == str(
        tmp_path / "test-item/20201211_223832_CS2.jpg"
    )


async def test_download_collection(tmp_path: Path, collection: Collection) -> None:
    collection = await stac_asset.download_collection(
        collection, tmp_path, file_name="collection.json"
    )
    assert os.path.exists(tmp_path / "collection.json")
    assert os.path.exists(tmp_path / "20201211_223832_CS2.jpg")
    asset = collection.assets["data"]
    assert asset.href == "./20201211_223832_CS2.jpg"


async def test_item_download_no_directory(tmp_path: Path, item: Item) -> None:
    with pytest.raises(DownloadError):
        await stac_asset.download_item(
            item, tmp_path / "doesnt-exist", config=Config(make_directory=False)
        )


async def test_item_download_key(tmp_path: Path, item: Item) -> None:
    await stac_asset.download_item(
        item, tmp_path, config=Config(file_name_strategy=FileNameStrategy.KEY)
    )
    assert Path(tmp_path / "data.jpg").exists()


async def test_item_download_same_file_name(tmp_path: Path, item: Item) -> None:
    item.assets["other-data"] = item.assets["data"].clone()
    with pytest.raises(AssetOverwriteError):
        await stac_asset.download_item(item, tmp_path)


async def test_include(tmp_path: Path, item: Item) -> None:
    item.assets["other-data"] = item.assets["data"].clone()
    item = await stac_asset.download_item(
        item, tmp_path, config=Config(include=["data"])
    )
    assert len(item.assets) == 1


async def test_exclude(tmp_path: Path, item: Item) -> None:
    item.assets["other-data"] = item.assets["data"].clone()
    await stac_asset.download_item(
        item, tmp_path, config=Config(exclude=["other-data"])
    )


async def test_cant_include_and_exclude(tmp_path: Path, item: Item) -> None:
    item.assets["other-data"] = item.assets["data"].clone()
    with pytest.raises(ConfigError):
        await stac_asset.download_item(
            item, tmp_path, config=Config(include=["data"], exclude=["other-data"])
        )


@pytest.mark.network_access
async def test_multiple_clients(tmp_path: Path, item: Item) -> None:
    item.assets["remote"] = Asset(
        href="https://storage.googleapis.com/open-cogs/stac-examples/20201211_223832_CS2.jpg",
    )
    item = await stac_asset.download_item(
        item, tmp_path, config=Config(file_name_strategy=FileNameStrategy.KEY)
    )


@pytest.mark.network_access
async def test_preconfigured_clients(tmp_path: Path, item: Item) -> None:
    item.assets["remote"] = Asset(
        href="s3://usgs-landsat/collection02/level-2/standard/oli-tirs/2023/092/068/LC09_L2SP_092068_20230522_20230524_02_T2/LC09_L2SP_092068_20230522_20230524_02_T2_thumb_small.jpeg",
    )
    s3_client = S3Client(requester_pays=True)
    item = await stac_asset.download_item(item, tmp_path, clients=[s3_client])


async def test_queue(tmp_path: Path, item: Item) -> None:
    messages: MessageQueue = Queue()
    item = await stac_asset.download_item(item, tmp_path, messages=messages)
    assert not messages.empty()


async def test_asset_exists(item: Item) -> None:
    assert await stac_asset.asset_exists(item.assets["data"])
    assert not await stac_asset.asset_exists(Asset(href="not-a-file"))


async def test_assert_asset_exists(item: Item) -> None:
    await stac_asset.assert_asset_exists(item.assets["data"])
    with pytest.raises(ValueError):
        await stac_asset.assert_asset_exists(Asset(href="not-a-file"))


async def test_open_href(data_path: Path) -> None:
    text = b""
    async for chunk in stac_asset.open_href(str(data_path / "item.json")):
        text += chunk
    Item.from_dict(json.loads(text))


async def test_read_href(data_path: Path) -> None:
    text = await stac_asset.read_href(str(data_path / "item.json"))
    Item.from_dict(json.loads(text))


async def test_keep_non_downloaded(item: Item, tmp_path: Path) -> None:
    item.assets["other-data"] = item.assets["data"].clone()
    item = await stac_asset.download_item(
        item, tmp_path, config=Config(include=["data"]), keep_non_downloaded=True
    )
    assert len(item.assets) == 2



from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from yarl import URL


@dataclass
class Message:
    """A message about downloading."""


@dataclass
class StartAssetDownload(Message):
    """Sent when an asset starts downloading."""

    key: str
    """The asset key."""

    owner_id: Optional[str]
    """The owner id."""

    href: str
    """The asset href."""

    path: Path
    """The local path that the asset is being downloaded to."""


@dataclass
class ErrorAssetDownload(Message):
    """Sent when an asset errors while downloading."""

    key: str
    """The asset key."""

    href: str
    """The asset href."""

    path: Path
    """The local path that the asset is being downloaded to."""

    error: Exception
    """The error."""


@dataclass
class SkipAssetDownload(Message):
    """Sent when an asset errors while downloading."""

    key: str
    """The asset key."""

    path: Path
    """The local path that the asset is being downloaded to."""


@dataclass
class FinishAssetDownload(Message):
    """Sent when an asset finishes downloading."""

    key: str
    """The asset key."""

    href: str
    """The asset href."""

    path: Path
    """The local path that the asset is being downloaded to."""


@dataclass
class WriteChunk(Message):
    """Sent when a chunk is written to disk."""

    href: str
    """The asset href."""

    path: Path
    """The local path that the asset is being downloaded to."""

    size: int
    """The number of bytes written."""


@dataclass
class OpenUrl(Message):
    """Sent when a url is first opened."""

    url: URL
    """The URL"""

    size: Optional[int]
    """The file size."""



import pytest
from stac_asset import ContentTypeError, validate


def test_content_type() -> None:
    validate.content_type("foo", "foo")
    with pytest.raises(ContentTypeError):
        validate.content_type("foo", "bar")
    validate.content_type(
        "image/tiff", "image/tiff; application=geotiff; profile=cloud-optimized"
    )
    validate.content_type(
        "image/tiff; application=geotiff; profile=cloud-optimized", "image/tiff"
    )
    validate.content_type("binary/octet-stream", "doesn't matter")
    validate.content_type("application/octet-stream", "doesn't matter")



from asyncio import Queue
from os import PathLike
from typing import Any, Union

from .messages import Message

MessageQueue = Queue[Message]

PathLikeObject = Union[PathLike[Any], str]
"""An object representing a file system path, except we exclude `bytes` because
`Path()` doesn't accept `bytes`.

A path-like object is either a str or bytes object representing a path, or an
object implementing the os.PathLike protocol. An object that supports the
os.PathLike protocol can be converted to a str or bytes file system path by
calling the os.fspath() function; os.fsdecode() and os.fsencode() can be used to
guarantee a str or bytes result instead, respectively. Introduced by PEP 519.

https://docs.python.org/3/glossary.html#term-path-like-object
"""



from enum import Enum, auto


class FileNameStrategy(Enum):
    """Strategy to use for naming files."""

    FILE_NAME = auto()
    """Save the asset with the file name in its href.

    Could potentially conflict with another asset with the same file name but
    different path.
    """

    KEY = auto()
    """Save the asset with its key as its file name."""


class ErrorStrategy(Enum):
    """Strategy to use when encountering errors during download."""

    KEEP = auto()
    """Keep the asset on the item with its original href."""

    DELETE = auto()
    """Delete the asset from the item."""



from __future__ import annotations

import asyncio
import json
import os.path
import warnings
from asyncio import Task
from dataclasses import dataclass
from pathlib import Path
from types import TracebackType
from typing import (
    AsyncIterator,
    List,
    Optional,
    Set,
    Type,
    Union,
)

import pystac.utils
from pystac import Asset, Collection, Item, ItemCollection, STACError
from yarl import URL

from .client import Client, Clients
from .config import Config
from .errors import AssetOverwriteError, DownloadError, DownloadWarning
from .messages import (
    ErrorAssetDownload,
    FinishAssetDownload,
    SkipAssetDownload,
    StartAssetDownload,
)
from .strategy import ErrorStrategy, FileNameStrategy
from .types import MessageQueue, PathLikeObject


@dataclass
class Download:
    owner: Union[Item, Collection]
    key: str
    asset: Asset
    path: Path
    clients: Clients
    config: Config

    async def download(
        self,
        messages: Optional[MessageQueue],
    ) -> Union[Download, WrappedError]:
        if not os.path.exists(self.path) or self.config.overwrite:
            try:
                await download_asset(
                    self.key,
                    self.asset,
                    self.path,
                    config=self.config,
                    messages=messages,
                    clients=self.clients,
                )
            except Exception as error:
                if self.config.fail_fast:
                    raise error
                else:
                    return WrappedError(self, error)
        else:
            if messages:
                await messages.put(SkipAssetDownload(key=self.key, path=self.path))

        self.asset.href = str(self.path)
        return self


class Downloads:
    clients: Clients
    config: Config
    downloads: List[Download]

    def __init__(self, config: Config, clients: Optional[List[Client]] = None) -> None:
        config.validate()
        self.config = config
        self.downloads = list()
        self.clients = Clients(config, clients)

    async def add(
        self,
        stac_object: Union[Item, Collection],
        root: Path,
        file_name: Optional[str],
        keep_non_downloaded: bool,
    ) -> None:
        stac_object = make_link_hrefs_absolute(stac_object)
        # Will fail if the stac object doesn't have a self href and there's
        # relative asset hrefs
        stac_object = make_asset_hrefs_absolute(stac_object)
        if file_name:
            stac_object.set_self_href(str(Path(root) / file_name))
        else:
            stac_object.set_self_href(None)

        asset_file_names: Set[str] = set()
        assets = dict()
        for key, asset in (
            (k, a)
            for k, a in stac_object.assets.items()
            if (not self.config.include or k in self.config.include)
            and (not self.config.exclude or k not in self.config.exclude)
        ):
            if self.config.file_name_strategy == FileNameStrategy.FILE_NAME:
                asset_file_name = os.path.basename(URL(asset.href).path)
            elif self.config.file_name_strategy == FileNameStrategy.KEY:
                asset_file_name = key + Path(asset.href).suffix
            else:
                raise ValueError(
                    f"unexpected file name strategy: {self.config.file_name_strategy}"
                )
            if asset_file_name in asset_file_names:
                raise AssetOverwriteError(list(asset_file_names))

            asset_file_names.add(asset_file_name)
            assets[key] = asset
            self.downloads.append(
                Download(
                    owner=stac_object,
                    key=key,
                    asset=asset,
                    path=root / asset_file_name,
                    clients=self.clients,
                    config=self.config,
                )
            )
        if keep_non_downloaded:
            stac_object.assets.update(assets)
        else:
            stac_object.assets = assets

    async def download(self, messages: Optional[MessageQueue]) -> None:
        tasks: Set[Task[Union[Download, WrappedError]]] = set()
        for download in self.downloads:
            task = asyncio.create_task(
                download.download(
                    messages=messages,
                )
            )
            tasks.add(task)
            task.add_done_callback(tasks.discard)

        try:
            results = await asyncio.gather(*tasks)
        except Exception as error:
            # We failed fast
            for task in tasks:
                if not task.done():
                    task.cancel()
            await asyncio.gather(*tasks, return_exceptions=True)
            raise error

        exceptions = list()
        for result in results:
            if isinstance(result, WrappedError):
                if self.config.error_strategy == ErrorStrategy.DELETE:
                    del result.download.owner.assets[result.download.key]
                else:
                    # Simple check to make sure we haven't added other
                    # strategies that we're not handling
                    assert self.config.error_strategy == ErrorStrategy.KEEP

                if self.config.warn:
                    warnings.warn(str(result.error), DownloadWarning)
                else:
                    exceptions.append(result.error)
        if exceptions:
            raise DownloadError(exceptions)

    async def __aenter__(self) -> Downloads:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.clients.close_all()


class WrappedError:
    download: Download
    error: Exception

    def __init__(self, download: Download, error: Exception) -> None:
        self.download = download
        self.error = error


async def download_item(
    item: Item,
    directory: PathLikeObject,
    file_name: Optional[str] = None,
    infer_file_name: bool = True,
    config: Optional[Config] = None,
    messages: Optional[MessageQueue] = None,
    clients: Optional[List[Client]] = None,
    keep_non_downloaded: bool = False,
) -> Item:
    """Downloads an item to the local filesystem.

    Args:
        item: The :py:class:`pystac.Item`.
        directory: The output directory that will hold the items and assets.
        file_name: The name of the item file to save. If not provided, will not
            be saved.
        infer_file_name: If ``file_name`` is None, infer the file name from the
            item's id. This argument is unused if ``file_name`` is not None.
        config: The download configuration
        messages: An optional queue to use for progress reporting
        clients: Pre-configured clients to use for access
        keep_non_downloaded: Keep all assets on the item, even if they're not
            downloaded.

    Returns:
        Item: The `~pystac.Item`, with the updated asset hrefs and self href.

    Raises:
        ValueError: Raised if the item doesn't have any assets.
    """
    if file_name is None and infer_file_name:
        file_name = f"{item.id}.json"

    async with Downloads(
        config=config or Config(),
        clients=clients,
    ) as downloads:
        await downloads.add(item, Path(directory), file_name, keep_non_downloaded)
        await downloads.download(messages)

    self_href = item.get_self_href()
    if self_href:
        make_asset_hrefs_relative(item)
        d = item.to_dict(include_self_link=True, transform_hrefs=False)
        os.makedirs(os.path.dirname(self_href), exist_ok=True)
        with open(self_href, "w") as f:
            json.dump(d, f)

    return item


async def download_collection(
    collection: Collection,
    directory: PathLikeObject,
    file_name: Optional[str] = "collection.json",
    config: Optional[Config] = None,
    messages: Optional[MessageQueue] = None,
    clients: Optional[List[Client]] = None,
    keep_non_downloaded: bool = False,
) -> Collection:
    """Downloads a collection to the local filesystem.

    Does not download the collection's items' assets -- use
    :py:func:`download_item_collection` to download multiple items.

    Args:
        collection: A pystac collection
        directory: The destination directory
        file_name: The name of the collection file to save. If not provided,
            will not be saved.
        config: The download configuration
        messages: An optional queue to use for progress reporting
        clients: Pre-configured clients to use for access
        keep_non_downloaded: Keep all assets on the item, even if they're not
            downloaded.

    Returns:
        Collection: The collection, with updated asset hrefs

    Raises:
        CantIncludeAndExclude: Raised if both include and exclude are not None.
    """
    async with Downloads(config=config or Config(), clients=clients) as downloads:
        await downloads.add(collection, Path(directory), file_name, keep_non_downloaded)
        await downloads.download(messages)

    self_href = collection.get_self_href()
    if self_href:
        make_asset_hrefs_relative(collection)
        d = collection.to_dict(include_self_link=True, transform_hrefs=False)
        with open(self_href, "w") as f:
            json.dump(d, f)

    return collection


async def download_item_collection(
    item_collection: ItemCollection,
    directory: PathLikeObject,
    file_name: Optional[str] = "item-collection.json",
    config: Optional[Config] = None,
    messages: Optional[MessageQueue] = None,
    clients: Optional[List[Client]] = None,
    keep_non_downloaded: bool = False,
) -> ItemCollection:
    """Downloads an item collection to the local filesystem.

    Args:
        item_collection: The item collection to download
        directory: The destination directory
        file_name: The name of the item collection file to save. If not
            provided, will not be saved.
        config: The download configuration
        messages: An optional queue to use for progress reporting
        clients: Pre-configured clients to use for access
        keep_non_downloaded: Keep all assets on the item, even if they're not
            downloaded.

    Returns:
        ItemCollection: The item collection, with updated asset hrefs

    Raises:
        CantIncludeAndExclude: Raised if both include and exclude are not None.
    """
    async with Downloads(config=config or Config(), clients=clients) as downloads:
        for item in item_collection.items:
            item.set_self_href(None)
            root = Path(directory) / item.id
            await downloads.add(item, root, None, keep_non_downloaded)
        await downloads.download(messages)
    if file_name:
        dest_href = Path(directory) / file_name
        for item in item_collection.items:
            for asset in item.assets.values():
                asset.href = pystac.utils.make_absolute_href(
                    asset.href, start_is_dir=True
                )
        item_collection.save_object(dest_href=str(dest_href))

    return item_collection


async def download_asset(
    key: str,
    asset: Asset,
    path: Path,
    config: Config,
    messages: Optional[MessageQueue] = None,
    clients: Optional[Clients] = None,
) -> Asset:
    """Downloads an asset.

    Args:
        key: The asset key
        asset: The asset
        path: The path to which the asset will be downloaded
        config: The download configuration
        messages: An optional queue to use for progress reporting
        clients: A async-safe cache of clients. If not provided, a new one
            will be created.

    Returns:
        Asset: The asset with an updated href

    Raises:
        ValueError: Raised if the asset does not have an absolute href
    """
    if clients is None:
        clients = Clients(config)

    if not path.parent.exists():
        if config.make_directory:
            path.parent.mkdir(parents=True, exist_ok=True)
        else:
            raise FileNotFoundError(f"output directory does not exist: {path.parent}")

    href = get_absolute_asset_href(
        asset=asset, alternate_assets=config.alternate_assets
    )
    if href is None:
        raise ValueError(f"asset '{key}' does not have an absolute href: {asset.href}")
    client = await clients.get_client(href)

    if messages:
        if asset.owner:
            assert isinstance(asset.owner, (Item, Collection))
            owner_id = asset.owner.id
        else:
            owner_id = None
        await messages.put(
            StartAssetDownload(key=key, href=href, path=path, owner_id=owner_id)
        )
    try:
        await client.download_href(
            href,
            path,
            clean=config.clean,
            content_type=asset.media_type,
            messages=messages,
        )
    except Exception as error:
        if messages:
            await messages.put(
                ErrorAssetDownload(key=key, href=href, path=path, error=error)
            )
        raise error

    asset.href = str(path)
    if messages:
        await messages.put(FinishAssetDownload(key=key, href=href, path=path))
    return asset


async def assert_asset_exists(
    asset: Asset,
    config: Optional[Config] = None,
    clients: Optional[List[Client]] = None,
) -> None:
    """Asserts that an asset exists.

    Raises the source error if it does not.

    Args:
        asset: The asset the check for existence
        config: The download configuration to use for the existence check
        clients: Any pre-configured clients to use for the existence check

    Raises:
        Exception: An exception from the underlying client.
    """
    if config is None:
        config = Config()
    clients_ = Clients(config, clients=clients)
    href = get_absolute_asset_href(asset, config.alternate_assets)
    if href:
        client = await clients_.get_client(href)
        await client.assert_href_exists(href)
    else:
        raise ValueError("asset does not have an absolute href")


async def asset_exists(
    asset: Asset,
    config: Optional[Config] = None,
    clients: Optional[List[Client]] = None,
) -> bool:
    """Returns true if an asset exists.

    Args:
        asset: The asset the check for existence
        config: The download configuration to use for the existence check
        clients: Any pre-configured clients to use for the existence check

    Returns:
        bool: Whether the asset exists or not
    """
    try:
        await assert_asset_exists(asset, config, clients)
    except Exception:
        return False
    else:
        return True


async def open_href(
    href: str, config: Optional[Config] = None, clients: Optional[List[Client]] = None
) -> AsyncIterator[bytes]:
    """Opens an href and yields byte chunks.

    Args:
        href: The href to read
        config: The download configuration to use
        clients: Any pre-configured clients to use

    Yields:
        bytes: The bytes from the href
    """
    if config is None:
        config = Config()
    clients_ = Clients(config, clients=clients)
    async with await clients_.get_client(href) as client:
        async for chunk in client.open_href(href):
            yield chunk


async def read_href(
    href: str, config: Optional[Config] = None, clients: Optional[List[Client]] = None
) -> bytes:
    """Reads an href and returns its bytes.

    Args:
        href: The href to read
        config: The download configuration to use
        clients: Any pre-configured clients to use

    Returns:
        bytes: The bytes from the href
    """
    data = b""
    async for chunk in open_href(href, config=config, clients=clients):
        data += chunk
    return data


def make_asset_hrefs_relative(
    stac_object: Union[Item, Collection],
) -> Union[Item, Collection]:
    # Copied from
    # https://github.com/stac-utils/pystac/blob/381cf89fc25c15142fb5a187d905e22681de42a2/pystac/item.py#L284C5-L298C20
    # until a fix for https://github.com/stac-utils/pystac/issues/1199 is
    # released.
    self_href = stac_object.get_self_href()
    for asset in stac_object.assets.values():
        if pystac.utils.is_absolute_href(asset.href):
            if self_href is None:
                raise STACError(
                    "Cannot make asset HREFs relative " "if no self_href is set."
                )
            asset.href = pystac.utils.make_relative_href(asset.href, self_href)
    return stac_object


def make_asset_hrefs_absolute(
    stac_object: Union[Item, Collection],
) -> Union[Item, Collection]:
    # Copied from
    # https://github.com/stac-utils/pystac/blob/381cf89fc25c15142fb5a187d905e22681de42a2/pystac/item.py#L309C3-L319C1
    # until a fix for https://github.com/stac-utils/pystac/issues/1199 is
    # released.
    self_href = stac_object.get_self_href()
    for asset in stac_object.assets.values():
        if not pystac.utils.is_absolute_href(asset.href):
            if self_href is None:
                raise STACError(
                    "Cannot make asset HREFs absolute if no self_href is set."
                )
            asset.href = pystac.utils.make_absolute_href(asset.href, self_href)
    return stac_object


def make_link_hrefs_absolute(
    stac_object: Union[Item, Collection], drop: bool = True
) -> Union[Item, Collection]:
    # This could be in pystac w/ STACObject as the input+output type
    links = list()
    for link in stac_object.links:
        absolute_href = link.get_absolute_href()
        if absolute_href:
            link.target = absolute_href
            links.append(link)
        elif not drop:
            raise ValueError(f"cannot make link's href absolute: {link}")
    stac_object.links = links
    return stac_object


def get_absolute_asset_href(asset: Asset, alternate_assets: List[str]) -> Optional[str]:
    alternate = asset.extra_fields.get("alternate")
    if not isinstance(alternate, dict):
        alternate = None
    if alternate and alternate_assets:
        for alternate_asset in alternate_assets:
            if alternate_asset in alternate:
                try:
                    href = alternate[alternate_asset]["href"]
                    if asset.owner:
                        start_href = asset.owner.get_self_href()
                    else:
                        start_href = None
                    return pystac.utils.make_absolute_href(
                        href, start_href, start_is_dir=False
                    )
                except KeyError:
                    raise ValueError(
                        "invalid alternate asset definition (missing href): "
                        f"{alternate}"
                    )
    return asset.get_absolute_href()



from __future__ import annotations

import os
from types import TracebackType
from typing import Optional, Type

from aiohttp import ClientSession

from .config import Config
from .http_client import HttpClient


class EarthdataClient(HttpClient):
    """Access data from https://www.earthdata.nasa.gov/.

    To access data, you'll need a personal access token.

    1. Create a new personal access token by going to
       https://urs.earthdata.nasa.gov/profile and then clicking "Generate Token"
       (you'll need to log in).
    2. Set an environment variable named ``EARTHDATA_PAT`` to your token.
    3. Use :py:meth:`EarthdataClient.from_config()` to create a new client.

    You can also provide your token directly to :py:func:`EarthdataClient.login()`.
    """

    @classmethod
    async def from_config(cls, config: Config) -> EarthdataClient:
        """Logs in to Earthdata and returns the default earthdata client.

        Uses a token stored in the ``EARTHDATA_PAT`` environment variable, if
        the token is not provided in the config.

        Args:
            config: A configuration object.

        Returns:
            EarthdataClient: A logged-in EarthData client.
        """
        return await cls.login(config.earthdata_token)

    @classmethod
    async def login(cls, token: Optional[str] = None) -> EarthdataClient:
        """Logs in to Earthdata and returns a client.

        If token is not provided, it is read from the ``EARTHDATA_PAT``
        environment variable.

        Args:
            token: The Earthdata bearer token

        Returns:
            EarthdataClient: A client configured to use the bearer token
        """
        if token is None:
            try:
                token = os.environ["EARTHDATA_PAT"]
            except KeyError:
                raise ValueError(
                    "token was not provided, and EARTHDATA_PAT environment variable "
                    "not set"
                )
        session = ClientSession(headers={"Authorization": f"Bearer {token}"})
        return cls(session)

    async def __aenter__(self) -> EarthdataClient:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Optional[bool]:
        await self.close()
        return await super().__aexit__(exc_type, exc_val, exc_tb)



from __future__ import annotations

import datetime
from asyncio import Lock, Queue
from datetime import timezone
from types import TracebackType
from typing import Any, AsyncIterator, Dict, Optional, Type

import dateutil.parser
from aiohttp import ClientSession
from yarl import URL

from .http_client import HttpClient

DEFAULT_SAS_TOKEN_ENDPOINT = "https://planetarycomputer.microsoft.com/api/sas/v1/token"


class _Token:
    expiry: datetime.datetime
    token: str

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> _Token:
        try:
            expiry = dateutil.parser.isoparse(data["msft:expiry"])
        except KeyError:
            raise ValueError(f"missing 'msft:expiry' key in dict: {data}")

        try:
            token = data["token"]
        except KeyError:
            raise ValueError(f"missing 'token' key in dict: {data}")

        return cls(expiry=expiry, token=token)

    def __init__(self, expiry: datetime.datetime, token: str) -> None:
        self.expiry = expiry
        self.token = token

    def ttl(self) -> float:
        return (self.expiry - datetime.datetime.now(timezone.utc)).total_seconds()

    def __str__(self) -> str:
        return self.token


class PlanetaryComputerClient(HttpClient):
    """Open and download assets from Microsoft's Planetary Computer.

    Heavily cribbed from
    https://github.com/microsoft/planetary-computer-sdk-for-python/blob/main/planetary_computer/sas.py,
    thanks Tom Augspurger!
    """

    _cache: Dict[URL, _Token]
    _cache_lock: Lock
    token_request_url: URL

    def __init__(
        self,
        session: ClientSession,
        sas_token_endpoint: str = DEFAULT_SAS_TOKEN_ENDPOINT,
    ) -> None:
        super().__init__(session)
        self._cache = dict()
        self._cache_lock = Lock()
        self.sas_token_endpoint = URL(sas_token_endpoint)

    async def open_url(
        self,
        url: URL,
        content_type: Optional[str] = None,
        messages: Optional[Queue[Any]] = None,
    ) -> AsyncIterator[bytes]:
        """Opens a url and iterates over its bytes.

        Includes functionality to sign the url with a SAS token fetched from
        this client's ``sas_token_endpoint``. Tokens are cached on a per-client
        basis to prevent a large number of requests when fetching many assets.

        Not every URL is modified with a SAS token. We only modify the url if:

        - The url is in Azure blob storage
        - The url is not in the public thumbnail storage account
        - The url hasn't already signed (we check this by seeing if the url has
            SAS-like query parameters)

        Args:
            url: The url to open
            content_type: The expected content type
            messages: An optional queue to use for progress reporting

        Yields:
            AsyncIterator[bytes]: An iterator over the file's bytes
        """
        url = await self._maybe_sign_url(url)
        async for chunk in super().open_url(
            url, content_type=content_type, messages=messages
        ):
            yield chunk

    async def assert_href_exists(self, href: str) -> None:
        """Asserts that the href exists.

        Uses a HEAD request on a signed url.
        """
        href = await self._maybe_sign_href(href)
        async with self.session.head(href) as response:
            response.raise_for_status()

    async def _sign(self, url: URL) -> URL:
        assert url.host
        account_name = url.host.split(".")[0]
        container_name = url.path.split("/", 2)[1]
        token = await self._get_token(account_name, container_name)
        return URL(str(url.with_query(None)) + "?" + token, encoded=False)

    async def _maybe_sign_url(self, url: URL) -> URL:
        if (
            url.host is not None
            and url.host.endswith(".blob.core.windows.net")
            and not url.host == "ai4edatasetspublicassets.blob.core.windows.net"
            and not set(url.query) & {"st", "se", "sp"}
        ):
            return await self._sign(url)
        else:
            return url

    async def _maybe_sign_href(self, href: str) -> str:
        return str(await self._maybe_sign_url(URL(href)))

    async def _get_token(self, account_name: str, container_name: str) -> str:
        url = self.sas_token_endpoint.joinpath(account_name, container_name)
        async with self._cache_lock:
            token = self._cache.get(url)
            if token is None or token.ttl() < 60:
                response = await self.session.get(url)
                response.raise_for_status()
                token = _Token.from_dict(await response.json())
                self._cache[url] = token
        return str(token)

    async def __aenter__(self) -> PlanetaryComputerClient:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Optional[bool]:
        await self.close()
        return await super().__aexit__(exc_type, exc_val, exc_tb)



import json
import os
from pathlib import Path

import pytest
import stac_asset._cli
from click.testing import CliRunner
from pystac import Item, ItemCollection


def test_download_item(tmp_path: Path, item_path: Path) -> None:
    runner = CliRunner()
    result = runner.invoke(
        stac_asset._cli.cli,
        ["download", str(item_path), str(tmp_path)],
    )
    assert result.exit_code == 0


def test_download_item_stdin_stdout(tmp_path: Path, item: Item) -> None:
    previous_working_directory = os.getcwd()
    os.chdir(tmp_path)
    try:
        item_as_str = json.dumps(
            item.to_dict(include_self_link=True, transform_hrefs=False)
        )
        runner = CliRunner(mix_stderr=False)
        result = runner.invoke(stac_asset._cli.cli, ["download"], input=item_as_str)
        assert result.exit_code == 0, result.stdout
        Item.from_dict(json.loads(result.stdout))
    finally:
        os.chdir(previous_working_directory)


def test_download_item_collection_stdin_stdout(
    tmp_path: Path, item_collection_path: Path
) -> None:
    runner = CliRunner()
    result = runner.invoke(
        stac_asset._cli.cli,
        [
            "download",
            str(item_collection_path),
            str(tmp_path),
            "--file-name",
            "item-collection.json",
        ],
    )
    assert result.exit_code == 0, result.stdout
    ItemCollection.from_file(str(tmp_path / "item-collection.json"))


def test_download_item_collection_file_name(
    tmp_path: Path, item_collection: ItemCollection
) -> None:
    previous_working_directory = os.getcwd()
    os.chdir(tmp_path)
    try:
        item_collection_as_str = json.dumps(
            item_collection.to_dict(transform_hrefs=False)
        )
        runner = CliRunner(mix_stderr=False)
        result = runner.invoke(
            stac_asset._cli.cli, ["download"], input=item_collection_as_str
        )
        assert result.exit_code == 0, result.stdout
        ItemCollection.from_dict(json.loads(result.stdout))
    finally:
        os.chdir(previous_working_directory)


@pytest.mark.network_access
def test_download_item_s3_requester_pays(tmp_path: Path) -> None:
    runner = CliRunner()
    result = runner.invoke(
        stac_asset._cli.cli,
        [
            "download",
            "https://landsatlook.usgs.gov/stac-server/collections/landsat-c2l2-sr/items/LC09_L2SP_092068_20230607_20230609_02_T1_SR",
            str(tmp_path),
            "--s3-requester-pays",
            "-i",
            "thumbnail",
            "--alternate-assets",
            "s3",
        ],
    )
    assert result.exit_code == 0


def test_info(item_path: Path) -> None:
    runner = CliRunner()
    result = runner.invoke(
        stac_asset._cli.cli,
        ["info", str(item_path)],
    )
    assert result.exit_code == 0, result.stdout



import os.path
from pathlib import Path

import pytest
from stac_asset import FilesystemClient

pytestmark = pytest.mark.asyncio


@pytest.fixture
def asset_href() -> str:
    return str(Path(__file__).parent / "data" / "20201211_223832_CS2.jpg")


async def test_download(tmp_path: Path, asset_href: str) -> None:
    async with FilesystemClient() as client:
        await client.download_href(asset_href, tmp_path / "out.jpg")

    assert os.path.getsize(tmp_path / "out.jpg") == 31367


async def test_href_exists(asset_href: str) -> None:
    async with FilesystemClient() as client:
        assert await client.href_exists(asset_href)



import os.path
from pathlib import Path

import pytest
from stac_asset import Config, PlanetaryComputerClient

pytestmark = [
    pytest.mark.network_access,
    pytest.mark.asyncio,
]


@pytest.fixture
def asset_href() -> str:
    return "https://sentinel2l2a01.blob.core.windows.net/sentinel2-l2/48/X/VR/2023/05/24/S2B_MSIL2A_20230524T084609_N0509_R107_T48XVR_20230524T120352.SAFE/GRANULE/L2A_T48XVR_A032451_20230524T084603/QI_DATA/T48XVR_20230524T084609_PVI.tif"


async def test_download(tmp_path: Path, asset_href: str) -> None:
    async with await PlanetaryComputerClient.from_config(Config()) as client:
        await client.download_href(asset_href, tmp_path / "out.tif")

    assert os.path.getsize(tmp_path / "out.tif") == 4096


async def test_href_exists(tmp_path: Path, asset_href: str) -> None:
    async with await PlanetaryComputerClient.from_config(Config()) as client:
        assert await client.href_exists(asset_href)



from __future__ import annotations

from types import TracebackType
from typing import Any, AsyncIterator, Dict, Optional, Type

import aiobotocore.session
import botocore.config
from aiobotocore.session import AioSession, ClientCreatorContext
from botocore import UNSIGNED
from yarl import URL

from . import validate
from .client import Client
from .config import (
    DEFAULT_S3_MAX_ATTEMPTS,
    DEFAULT_S3_REGION_NAME,
    DEFAULT_S3_RETRY_MODE,
    Config,
)
from .messages import OpenUrl
from .types import MessageQueue


class S3Client(Client):
    """A client for interacting with s3 urls.

    To use the ``requester_pays`` option, you need to configure your AWS
    credentials. See `the AWS documentation
    <https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-files.html>`_
    for instructions.
    """

    session: AioSession
    """The session that will be used for all s3 requests."""

    region_name: str
    """The region that all clients will be rooted in."""

    requester_pays: bool
    """If True, enable access to `requester pays buckets
    <https://docs.aws.amazon.com/AmazonS3/latest/userguide/RequesterPaysBuckets.html>`_."""

    retry_mode: str
    """The retry mode, one of "adaptive", "legacy", or "standard".

    See `the boto3 docs
    <https://boto3.amazonaws.com/v1/documentation/api/latest/guide/retries.html>`_
    for more information on the available modes.
    """

    max_attempts: int
    """The maximum number of attempts."""

    @classmethod
    async def from_config(cls, config: Config) -> S3Client:
        """Creates an s3 client from a config.

        Args:
            config: The config object

        Returns:
            S3Client: A new s3 client
        """
        return cls(
            requester_pays=config.s3_requester_pays,
            region_name=config.s3_region_name,
            retry_mode=config.s3_retry_mode,
            max_attempts=config.s3_max_attempts,
        )

    def __init__(
        self,
        requester_pays: bool = False,
        region_name: str = DEFAULT_S3_REGION_NAME,
        retry_mode: str = DEFAULT_S3_RETRY_MODE,
        max_attempts: int = DEFAULT_S3_MAX_ATTEMPTS,
    ) -> None:
        super().__init__()
        self.session = aiobotocore.session.get_session()
        self.region_name = region_name
        self.requester_pays = requester_pays
        self.retry_mode = retry_mode
        self.max_attempts = max_attempts

    async def open_url(
        self,
        url: URL,
        content_type: Optional[str] = None,
        messages: Optional[MessageQueue] = None,
    ) -> AsyncIterator[bytes]:
        """Opens an s3 url and iterates over its bytes.

        Args:
            url: The url to open
            content_type: The expected content type
            messages: An optional queue to use for progress reporting

        Yields:
            AsyncIterator[bytes]: An iterator over the file's bytes

        Raises:
            SchemeError: Raised if the url's scheme is not ``s3``
        """
        async with self._create_client() as client:
            response = await client.get_object(**self._params(url))
            if content_type:
                validate.content_type(response["ContentType"], content_type)
            if messages:
                await messages.put(OpenUrl(url=url, size=response["ContentLength"]))
            async for chunk in response["Body"]:
                yield chunk

    async def has_credentials(self) -> bool:
        """Returns true if the sessions has credentials."""
        return await self.session.get_credentials() is not None

    async def assert_href_exists(self, href: str) -> None:
        """Asserts that the href exists.

        Uses ``head_object``
        """
        async with self._create_client() as client:
            await client.head_object(**self._params(URL(href)))

    def _create_client(self) -> ClientCreatorContext:
        retries = {
            "max_attempts": self.max_attempts,
            "mode": self.retry_mode,
        }
        if self.requester_pays:
            config = botocore.config.Config(retries=retries)
        else:
            config = botocore.config.Config(signature_version=UNSIGNED, retries=retries)
        return self.session.create_client(
            "s3", region_name=self.region_name, config=config
        )

    def _params(self, url: URL) -> Dict[str, Any]:
        bucket = url.host
        key = url.path[1:]
        params = {
            "Bucket": bucket,
            "Key": key,
        }
        if self.requester_pays:
            params["RequestPayer"] = "requester"
        return params

    async def __aenter__(self) -> S3Client:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Optional[bool]:
        return None



from __future__ import annotations

from types import TracebackType
from typing import AsyncIterator, Optional, Type, TypeVar

from aiohttp import ClientSession
from yarl import URL

from . import validate
from .client import Client
from .config import Config
from .messages import OpenUrl
from .types import MessageQueue

T = TypeVar("T", bound="HttpClient")


class HttpClient(Client):
    """A simple client for making HTTP requests.

    By default, doesn't do any authentication.
    Configure the session to customize its behavior.
    """

    check_content_type: bool
    """If true, check the asset's content type against the response from the server.

    See :py:func:`stac_asset.validate.content_type` for more information about
    hte content type check.
    """

    session: ClientSession
    """A atiohttp session that will be used for all requests."""

    @classmethod
    async def from_config(cls: Type[T], config: Config) -> T:
        """Creates the default http client with a vanilla session object."""
        # TODO add basic auth
        session = ClientSession()
        return cls(session)

    def __init__(self, session: ClientSession, check_content_type: bool = True) -> None:
        super().__init__()
        self.session = session
        self.check_content_type = check_content_type

    async def open_url(
        self,
        url: URL,
        content_type: Optional[str] = None,
        messages: Optional[MessageQueue] = None,
    ) -> AsyncIterator[bytes]:
        """Opens a url with this client's session and iterates over its bytes.

        Args:
            url: The url to open
            content_type: The expected content type
            messages: An optional queue to use for progress reporting

        Yields:
            AsyncIterator[bytes]: An iterator over the file's bytes

        Raises:
            :py:class:`aiohttp.ClientResponseError`: Raised if the response is not OK
        """
        async with self.session.get(url, allow_redirects=True) as response:
            response.raise_for_status()
            if self.check_content_type and content_type:
                validate.content_type(
                    actual=response.content_type, expected=content_type
                )
            if messages:
                await messages.put(OpenUrl(url=url, size=response.content_length))
            async for chunk, _ in response.content.iter_chunks():
                yield chunk

    async def assert_href_exists(self, href: str) -> None:
        """Asserts that the href exists.

        Uses a HEAD request.
        """
        async with self.session.head(href) as response:
            response.raise_for_status()

    async def close(self) -> None:
        """Close this http client.

        Closes the underlying session.
        """
        await self.session.close()

    async def __aenter__(self) -> HttpClient:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Optional[bool]:
        await self.close()
        return await super().__aexit__(exc_type, exc_val, exc_tb)
